@@ -11,7 +11,7 @@ import logging
 import time
 import typing
 
-from frozendict import frozendict
+from immutabledict import immutabledict
 import numpy as np
 
 import strax
@@ -46,9 +46,9 @@ class Plugin:
     """
     __version__ = '0.0.0'
 
-    # For multi-output plugins these should be (frozen)dicts
-    data_kind: typing.Union[str, frozendict, dict]
-    dtype: typing.Union[tuple, np.dtype, frozendict, dict]
+    # For multi-output plugins these should be (immutable)dicts
+    data_kind: typing.Union[str, immutabledict, dict]
+    dtype: typing.Union[tuple, np.dtype, immutabledict, dict]
 
     depends_on: tuple
     provides: tuple
@@ -78,7 +78,7 @@ class Plugin:
     # Do not specify attributes below
 
     # Set using the takes_config decorator
-    takes_config = frozendict()
+    takes_config = immutabledict()
 
     # These are set on plugin initialization, which is done in the core
     run_id: str
@@ -120,7 +120,7 @@ class Plugin:
         if self.multi_output:
             # Convert to a dict of numpy dtypes
             if (not hasattr(self, 'data_kind')
-                    or not isinstance(self.data_kind, dict)):
+                    or not isinstance(self.data_kind, (dict, immutabledict))):
                 raise ValueError(
                     f"{self.__class__.__name__} has multiple outputs and "
                     "must declare its data kind as a dict: "
@@ -228,6 +228,30 @@ class Plugin:
         # Don't raise NotImplementedError, IDE complains
         raise RuntimeError("source_finished called on a regular plugin")
 
+    def _fetch_chunk(self, d, iters, check_end_not_before=None):
+        """Add a chunk of the datatype d to the input buffer.
+        Return True if this succeeded, False if the source is exhausted.
+        :param d: data type to fetch
+        :param iters: iterators that produce data
+        :param check_end_not_before: Raise a runtimeError if the source 
+        is exhausted, but the input buffer ends before this time.
+        """
+        try:
+            # print(f"Fetching {d} in {self}, hope to see {hope_to_see}")
+            self.input_buffer[d] = strax.Chunk.concatenate(
+                [self.input_buffer[d], next(iters[d])])
+            # print(f"Fetched {d} in {self}, "
+            #      f"now have {self.input_buffer[d]}")
+            return True
+        except StopIteration:
+            # print(f"Got StopIteration while fetching for {d} in {self}")
+            if (check_end_not_before is not None
+                    and self.input_buffer[d].end < check_end_not_before):
+                raise RuntimeError(
+                    f"Tried to get data until {check_end_not_before}, but {d} "
+                    f"ended prematurely at {self.input_buffer[d].end}")
+            return False
+
     def iter(self, iters, executor=None):
         """Iterate over dependencies and yield results
 
@@ -237,30 +261,18 @@ class Plugin:
         """
         pending_futures = []
         last_input_received = time.time()
-
         self.input_buffer = {d: None
                              for d in self.depends_on}
-        if len(self.depends_on):
-            pacemaker = self.depends_on[0]
-        else:
-            pacemaker = None
 
-        def _fetch_chunk(d, hope_to_see=None):
-            try:
-                # print(f"Fetching {d} in {self}, hope to see {hope_to_see}")
-                self.input_buffer[d] = strax.Chunk.concatenate(
-                    [self.input_buffer[d], next(iters[d])])
-                # print(f"Fetched {d} in {self}, "
-                #      f"now have {self.input_buffer[d]}")
-                return True
-            except StopIteration:
-                # print(f"Got StopIteration while fetching for {d} in {self}")
-                if (hope_to_see is not None
-                        and self.input_buffer[d].end < hope_to_see):
-                    raise RuntimeError(
-                        f"Tried to get data until {hope_to_see}, but {d} "
-                        f"ended prematurely at {self.input_buffer[d].end}")
-                return False
+        # Fetch chunks from all inputs. Whoever is the slowest becomes the
+        # pacemaker
+        pacemaker = None
+        _end = float('inf')
+        for d in self.depends_on:
+            self._fetch_chunk(d, iters)
+            if self.input_buffer[d].end < _end:
+                pacemaker = d
+                _end = self.input_buffer[d].end
 
         for chunk_i in itertools.count():
 
@@ -269,7 +281,7 @@ class Plugin:
                 if self.source_finished():
                     # Chunk_i does not exist. We are done.
                     print("Source finished!")
-                    self.cleanup(wait_for=pending_futures)
+                    self.cleanup(iters, wait_for=pending_futures)
                     return
 
                 if time.time() > last_input_received + self.input_timeout:
@@ -286,10 +298,13 @@ class Plugin:
             if pacemaker is None:
                 inputs_merged = dict()
             else:
-                # Fetch the pacemaker, to figure out when this chunk ends
-                if not _fetch_chunk(pacemaker):
-                    self.cleanup(wait_for=pending_futures)
-                    return
+                if chunk_i != 0:
+                    # Fetch the pacemaker, to figure out when this chunk ends
+                    # (don't do it for chunk 0, for which we already fetched)
+                    if not self._fetch_chunk(pacemaker, iters):
+                        # Source exhausted. Cleanup will do final checks.
+                        self.cleanup(iters, wait_for=pending_futures)
+                        return
                 this_chunk_end = self.input_buffer[pacemaker].end
 
                 inputs = dict()
@@ -298,7 +313,9 @@ class Plugin:
                     if d != pacemaker:
                         while (self.input_buffer[d] is None
                                or self.input_buffer[d].end < this_chunk_end):
-                            _fetch_chunk(d, hope_to_see=this_chunk_end)
+                            self._fetch_chunk(
+                                d, iters,
+                                check_end_not_before=this_chunk_end)
                     inputs[d], self.input_buffer[d] = \
                         self.input_buffer[d].split(
                             t=this_chunk_end,
@@ -308,16 +325,25 @@ class Plugin:
                 # trim the others too.
                 # In very hairy cases this can take multiple passes.
                 # TODO: can we optimize this, or code it more elegantly?
-                while True:
+                max_passes_left = 10
+                while max_passes_left > 0:
                     this_chunk_end = min([x.end for x in inputs.values()]
                                          + [this_chunk_end])
                     if len(set([x.end for x in inputs.values()])) <= 1:
                         break
                     for d in self.depends_on:
-                        inputs[d], self.input_buffer[d] = \
-                            self.input_buffer[d].split(
+                        inputs[d], back_to_buffer = \
+                            inputs[d].split(
                                 t=this_chunk_end,
                                 allow_early_split=True)
+                        self.input_buffer[d] = strax.Chunk.concatenate(
+                            [back_to_buffer, self.input_buffer[d]])
+                    max_passes_left -= 1
+                else:
+                    raise RuntimeError(
+                        f"{self} was unable to get time-consistent "
+                        f"inputs after ten passess. Inputs: \n{inputs}\n"
+                        f"Input buffer:\n{self.input_buffer}")
 
                 # Merge inputs of the same kind
                 inputs_merged = {
@@ -339,9 +365,23 @@ class Plugin:
 
         raise RuntimeError("This cannot happen.")
 
-    def cleanup(self, wait_for):
-        # By default we do NOT wait for the computation futures to complete;
-        # there is nothing to specific to be done when a plugin exits.
+    def cleanup(self,
+                iters: typing.Dict[str, typing.Iterable],
+                wait_for):
+        # The wait_for option is only used in child classes;
+        # A standard plugin doesn't need to do anything with the computation
+        # future results.
+
+        # Check all sources are exhausted.
+        # This is more than a check though -- it ensure the content of
+        # all sources are requested all the way (including the final
+        # Stopiteration), as required by lazy-mode processing requires
+        for d in iters.keys():
+            if self._fetch_chunk(d, iters):
+                raise RuntimeError(
+                    f"Plugin {d} terminated without fetching last {d}!")
+
+        # Check the input buffer is empty
         for d, buffer in self.input_buffer.items():
             if buffer is not None and len(buffer):
                 raise RuntimeError(
@@ -486,18 +526,11 @@ class OverlapWindowPlugin(Plugin):
         raise NotImplementedError
 
     def iter(self, iters, executor=None):
-        # Keep one chunk in reserve, since we have to do something special
-        # if we see the last chunk.
-        last_result = None
-        for x in super().iter(iters, executor=executor):
-            if last_result is not None:
-                yield last_result
-            last_result = x
+        yield from super().iter(iters, executor=executor)
 
-        if self.cached_results is not None and last_result is not None:
-            # Note we do this even if the cached_result is only emptiness,
-            # to make sure our final result ends at the right time.
-            yield strax.Chunk.concatenate([last_result, self.cached_results])
+        # Yield final results, kept at bay in fear of a new chunk
+        if self.cached_results is not None:
+            yield self.cached_results
 
     def do_compute(self, chunk_i=None, **kwargs):
         if not len(kwargs):
@@ -636,10 +669,11 @@ class MergeOnlyPlugin(Plugin):
 
 @export
 class ParallelSourcePlugin(Plugin):
-    """An plugin that inlines the computation of other plugins and saving
-    of their results.
+    """An plugin that inlines the computations of other plugins
+    and the saving of their results.
 
-    This evades data transfer (pickling and/or memory copy) penalties.
+    This evades data transfer (pickling and/or memory copy) penalties
+    while multiprocessing.
     """
     parallel = 'process'
 
@@ -688,30 +722,30 @@ class ParallelSourcePlugin(Plugin):
         # Inline savers that do not require rechunking
         savers = components.savers
         sub_savers = dict()
-        for d, p in sub_plugins.items():
-
-            if d not in savers:
-                continue
-            if p.rechunk_on_save:
-                # Case 3. has a saver we can't inline (this is checked later)
-                outputs_to_send.add(d)
-                continue
-
-            remaining_savers = []
-            for s_i, s in enumerate(savers[d]):
-                if not s.allow_fork:
-                    # Case 3 again, cannot inline saver
+        for p in sub_plugins.values():
+            for d in p.provides:
+                if d not in savers:
+                    continue
+                if p.rechunk_on_save:
+                    # Case 3. has a saver we can't inline
                     outputs_to_send.add(d)
-                    remaining_savers.append(s)
                     continue
-                if d not in sub_savers:
-                    sub_savers[d] = []
-                s.is_forked = True
-                sub_savers[d].append(s)
-            savers[d] = remaining_savers
 
-            if not len(savers[d]):
-                del savers[d]
+                remaining_savers = []
+                for s_i, s in enumerate(savers[d]):
+                    if not s.allow_fork:
+                        # Case 3 again, cannot inline saver
+                        outputs_to_send.add(d)
+                        remaining_savers.append(s)
+                        continue
+                    if d not in sub_savers:
+                        sub_savers[d] = []
+                    s.is_forked = True
+                    sub_savers[d].append(s)
+                savers[d] = remaining_savers
+
+                if not len(savers[d]):
+                    del savers[d]
 
         p = cls(depends_on=sub_plugins[start_from].depends_on)
         p.sub_plugins = sub_plugins
@@ -753,10 +787,7 @@ class ParallelSourcePlugin(Plugin):
             for output_name, p in self.sub_plugins.items():
                 if output_name in results:
                     continue
-                # Sorting deps since otherwise input field order depends on
-                # order in which computation happened, which might be bad?
-                deps = sorted(p.depends_on)
-                if any([d not in results for d in deps]):
+                if any([d not in results for d in p.depends_on]):
                     continue
                 compute_kwargs = dict(chunk_i=chunk_i)
 
@@ -802,10 +833,10 @@ class ParallelSourcePlugin(Plugin):
 
         return self._fix_output(results, start=r0.start, end=r0.end)
 
-    def cleanup(self, wait_for):
+    def cleanup(self, iters, wait_for):
         print(f"{self.__class__.__name__} exhausted. "
               f"Waiting for {len(wait_for)} pending futures.")
         for savers in self.sub_savers.values():
             for s in savers:
                 s.close(wait_for=wait_for)
-        super().cleanup(wait_for)
+        super().cleanup(iters, wait_for)
